@@ -55,6 +55,8 @@ typedef struct {
   gboolean multiple;
   ExternalWindow *external_parent;
 
+  GSList *files;
+
   int response;
   GSList *uris;
 
@@ -71,6 +73,7 @@ file_dialog_handle_free (gpointer data)
   g_clear_object (&handle->external_parent);
   g_object_unref (handle->dialog);
   g_object_unref (handle->request);
+  g_slist_free_full (handle->files, g_free);
   g_slist_free_full (handle->uris, g_free);
   g_hash_table_unref (handle->choices);
 
@@ -132,6 +135,38 @@ send_response (FileDialogHandle *handle)
 
   g_variant_builder_init (&opt_builder, G_VARIANT_TYPE_VARDICT);
 
+  if (handle->action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER && handle->uris)
+    {
+      g_autoptr(GFile) base_dir = g_file_new_for_uri (handle->uris->data);
+
+      g_slist_free_full (handle->uris, g_free);
+      handle->uris = NULL;
+
+      for (l = handle->files; l; l = l->next)
+        {
+          int uniqifier = 0;
+          const char *file_name = l->data;
+          g_autoptr(GFile) file = g_file_get_child (base_dir, file_name);
+
+          while (g_file_query_exists(file, NULL))
+            {
+              g_autofree char *base_name = g_file_get_basename (file);
+              g_auto(GStrv) parts = NULL;
+              g_autoptr(GString) unique_name = NULL;
+
+              parts = g_strsplit (base_name, ".", 2);
+
+              unique_name = g_string_new (parts[0]);
+              g_string_append_printf (unique_name, "(%i)", ++uniqifier);
+              if (parts[1] != NULL)
+                  g_string_append (unique_name, parts[1]);
+
+              file = g_file_get_child (base_dir, unique_name->str);
+            }
+          handle->uris = g_slist_append (handle->uris, g_file_get_uri (file));
+        }
+    }
+
   g_variant_builder_init (&uri_builder, G_VARIANT_TYPE_STRING_ARRAY);
   for (l = handle->uris; l; l = l->next)
     {
@@ -152,11 +187,16 @@ send_response (FileDialogHandle *handle)
                                               handle->invocation,
                                               handle->response,
                                               g_variant_builder_end (&opt_builder));
-  else
+  else if (handle->action == GTK_FILE_CHOOSER_ACTION_SAVE)
     xdp_impl_file_chooser_complete_save_file (handle->impl,
                                               handle->invocation,
                                               handle->response,
                                               g_variant_builder_end (&opt_builder));
+  else
+    xdp_impl_file_chooser_complete_save_files (handle->impl,
+                                               handle->invocation,
+                                               handle->response,
+                                               g_variant_builder_end (&opt_builder));
 
   file_dialog_handle_close (handle);
 }
@@ -297,11 +337,16 @@ handle_close (XdpImplRequest *object,
                                               handle->invocation,
                                               2,
                                               g_variant_builder_end (&opt_builder));
-  else
+  else if (handle->action == GTK_FILE_CHOOSER_ACTION_SAVE)
     xdp_impl_file_chooser_complete_save_file (handle->impl,
                                               handle->invocation,
                                               2,
                                               g_variant_builder_end (&opt_builder));
+  else
+    xdp_impl_file_chooser_complete_save_files (handle->impl,
+                                               handle->invocation,
+                                               2,
+                                               g_variant_builder_end (&opt_builder));
   file_dialog_handle_close (handle);
 
   if (handle->request->exported)
@@ -362,6 +407,8 @@ handle_open (XdpImplFileChooser *object,
   const char *current_name;
   const char *path;
   g_autoptr (GVariant) choices = NULL;
+  g_autoptr (GVariant) current_filter = NULL;
+  GSList *filters = NULL;
   GtkWidget *preview;
 
   method_name = g_dbus_method_invocation_get_method_name (invocation);
@@ -374,6 +421,11 @@ handle_open (XdpImplFileChooser *object,
       action = GTK_FILE_CHOOSER_ACTION_SAVE;
       multiple = FALSE;
     }
+  else if (strcmp (method_name, "SaveFiles") == 0)
+    {
+      action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
+      multiple = FALSE;
+    }
   else
     {
       action = GTK_FILE_CHOOSER_ACTION_OPEN;
@@ -385,7 +437,12 @@ handle_open (XdpImplFileChooser *object,
     modal = TRUE;
 
   if (!g_variant_lookup (arg_options, "accept_label", "&s", &accept_label))
-    accept_label = _("_Open");
+    {
+      if (strcmp (method_name, "OpenFile") == 0)
+        accept_label = _("_Open");
+      else
+        accept_label = _("_Save");
+    }
 
   cancel_label = _("_Cancel");
 
@@ -466,11 +523,57 @@ handle_open (XdpImplFileChooser *object,
           GtkFileFilter *filter;
 
           filter = gtk_file_filter_new_from_gvariant (variant);
+          filters = g_slist_append (filters, g_object_ref (filter));
           gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
           g_variant_unref (variant);
         }
       g_variant_iter_free (iter);
     }
+
+  if (g_variant_lookup (arg_options, "current_filter", "@(sa(us))", &current_filter))
+    {
+      g_autoptr (GtkFileFilter) filter = NULL;
+      const char *current_filter_name;
+
+      filter = gtk_file_filter_new_from_gvariant (current_filter);
+      current_filter_name = gtk_file_filter_get_name (filter);
+
+      if (!filters)
+        {
+          /* We are setting a single, unchangeable filter. */
+          gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), filter);
+        }
+      else
+        {
+          gboolean handled = FALSE;
+
+          /* We are trying to select the default filter from the list of
+           * filters. We want to naively take filter and pass it to
+           * gtk_file_chooser_set_filter(), but it's not good enough
+           * because GTK+ just compares filters by pointer value, so the
+           * pointer itself has to match. We'll use the heuristic that
+           * if two filters have the same name, they must be the same
+           * unless the application is very dumb.
+           */
+          for (GSList *l = filters; l; l = l->next)
+            {
+              GtkFileFilter *f = l->data;
+              const char *name = gtk_file_filter_get_name (f);
+
+              if (g_strcmp0 (name, current_filter_name) == 0)
+                {
+                  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), f);
+                  handled = TRUE;
+                  break;
+                }
+            }
+
+          if (!handled)
+            g_warning ("current file filter must be present in filters list when list is nonempty");
+        }
+    }
+  g_slist_free_full (filters, g_object_unref);
+
   if (strcmp (method_name, "SaveFile") == 0)
     {
       if (g_variant_lookup (arg_options, "current_name", "&s", &current_name))
@@ -484,6 +587,24 @@ handle_open (XdpImplFileChooser *object,
       if (g_variant_lookup (arg_options, "current_file", "^&ay", &path))
         gtk_file_chooser_select_filename (GTK_FILE_CHOOSER (dialog), path);
     }
+  else if (strcmp (method_name, "SaveFiles") == 0)
+    {
+      /* TODO: is this useful ?
+       * In a sandboxed situation, the current folder and current file
+       * are likely in the fuse filesystem
+       */
+      if (g_variant_lookup (arg_options, "current_folder", "^&ay", &path))
+        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), path);
+
+      if (g_variant_lookup (arg_options, "files", "aay", &iter))
+        {
+          char *file = NULL;
+          while (g_variant_iter_next (iter, "^ay", &file))
+            handle->files = g_slist_append (handle->files, file);
+
+          g_variant_iter_free (iter);
+        }
+    }
 
   g_object_unref (fake_parent);
 
@@ -496,7 +617,7 @@ handle_open (XdpImplFileChooser *object,
 
       extra = gtk_file_chooser_get_extra_widget (GTK_FILE_CHOOSER (dialog));
 
-      readonly = gtk_check_button_new_with_label ("Open files read-only");
+      readonly = gtk_check_button_new_with_label (_("Open files read-only"));
       gtk_widget_show (readonly);
 
       g_signal_connect (readonly, "toggled",
@@ -530,6 +651,7 @@ file_chooser_init (GDBusConnection *bus,
 
   g_signal_connect (helper, "handle-open-file", G_CALLBACK (handle_open), NULL);
   g_signal_connect (helper, "handle-save-file", G_CALLBACK (handle_open), NULL);
+  g_signal_connect (helper, "handle-save-files", G_CALLBACK (handle_open), NULL);
 
   if (!g_dbus_interface_skeleton_export (helper,
                                          bus,
