@@ -269,6 +269,87 @@ cursor_mode_to_gnome_cursor_mode (ScreenCastCursorMode cursor_mode)
   g_assert_not_reached ();
 }
 
+static gboolean
+gnome_screen_cast_session_record_window (GnomeScreenCastSession *gnome_screen_cast_session,
+                                         const guint64 id,
+                                         ScreenCastSelection *select,
+                                         GError **error)
+{
+  OrgGnomeMutterScreenCastSession *session_proxy =
+    gnome_screen_cast_session->proxy;
+  GVariantBuilder properties_builder;
+  GVariant *properties;
+  g_autofree char *stream_path = NULL;
+  GDBusConnection *connection;
+  OrgGnomeMutterScreenCastStream *stream_proxy;
+  GnomeScreenCastStream *stream;
+  GVariant *parameters;
+
+  g_variant_builder_init (&properties_builder, G_VARIANT_TYPE_VARDICT);
+  g_variant_builder_add (&properties_builder, "{sv}",
+                         "window-id",
+                         g_variant_new_uint64 (id));
+  if (select->cursor_mode)
+    {
+      uint32_t gnome_cursor_mode;
+
+      gnome_cursor_mode = cursor_mode_to_gnome_cursor_mode (select->cursor_mode);
+      g_variant_builder_add (&properties_builder, "{sv}",
+                             "cursor-mode",
+                             g_variant_new_uint32 (gnome_cursor_mode));
+    }
+  properties = g_variant_builder_end (&properties_builder);
+
+  if (!org_gnome_mutter_screen_cast_session_call_record_window_sync (session_proxy,
+                                                                     properties,
+                                                                     &stream_path,
+                                                                     NULL,
+                                                                     error))
+    return FALSE;
+
+  connection = g_dbus_proxy_get_connection (G_DBUS_PROXY (session_proxy));
+  stream_proxy =
+    org_gnome_mutter_screen_cast_stream_proxy_new_sync (connection,
+                                                        G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
+                                                        "org.gnome.Mutter.ScreenCast",
+                                                        stream_path,
+                                                        NULL,
+                                                        error);
+  if (!stream_proxy)
+    return FALSE;
+
+  stream = g_object_new (gnome_screen_cast_stream_get_type (), NULL);
+  stream->session = gnome_screen_cast_session;
+  stream->path = g_strdup (stream_path);
+  stream->proxy = stream_proxy;
+
+  parameters = org_gnome_mutter_screen_cast_stream_get_parameters (stream->proxy);
+  if (parameters)
+    {
+      if (g_variant_lookup (parameters, "position", "(ii)",
+                            &stream->x, &stream->y))
+        stream->has_position = TRUE;
+      if (g_variant_lookup (parameters, "size", "(ii)",
+                            &stream->width, &stream->height))
+        stream->has_size = TRUE;
+    }
+  else
+    {
+      g_warning ("Screen cast stream %s missing parameters",
+                 stream->path);
+    }
+
+  g_signal_connect (stream_proxy, "pipewire-stream-added",
+                    G_CALLBACK (on_pipewire_stream_added),
+                    stream);
+
+  gnome_screen_cast_session->streams =
+    g_list_prepend (gnome_screen_cast_session->streams, stream);
+  gnome_screen_cast_session->n_needed_stream_node_ids++;
+
+  return TRUE;
+}
+
 static gboolean
 gnome_screen_cast_session_record_monitor (GnomeScreenCastSession *gnome_screen_cast_session,
                                           const char *connector,
@@ -362,14 +443,17 @@ gnome_screen_cast_session_record_selections (GnomeScreenCastSession *gnome_scree
     {
       ScreenCastSourceType source_type;
       g_autofree char *key = NULL;
+      g_autoptr(GVariant) variant = NULL;
+      guint64 id;
 
-      g_variant_get (selection, "(us)",
+      g_variant_get (selection, "(u?)",
                      &source_type,
-                     &key);
+                     &variant);
 
       switch (source_type)
         {
         case SCREEN_CAST_SOURCE_TYPE_MONITOR:
+          key = g_variant_dup_string (variant, NULL);
           if (!gnome_screen_cast_session_record_monitor (gnome_screen_cast_session,
                                                          key,
                                                          select,
@@ -377,8 +461,12 @@ gnome_screen_cast_session_record_selections (GnomeScreenCastSession *gnome_scree
             return FALSE;
           break;
         case SCREEN_CAST_SOURCE_TYPE_WINDOW:
-          g_assert_not_reached ();
-          return FALSE;
+          id = g_variant_get_uint64 (variant);
+          if (!gnome_screen_cast_session_record_window (gnome_screen_cast_session,
+                                                        id,
+                                                        select,
+                                                        error))
+            return FALSE;
         }
     }
 
